import { escapeRegExp as escapeRe } from 'lodash';
import correctFilePath from '../utils/correct-file-path';
import escapeUserAgent from '../utils/escape-user-agent';

const DATE_FORMAT = 'YYYY-MM-DD';
const TIME_FORMAT = 'HH-mm-ss';

const SCRENSHOT_EXTENTION = 'png';

const ERRORS_FOLDER = 'errors';

const PLACEHOLDERS = {
    DATE:               '${DATE}',
    TIME:               '${TIME}',
    TEST_INDEX:         '${TEST_INDEX}',
    FILE_INDEX:         '${FILE_INDEX}',
    QUARANTINE_ATTEMPT: '${QUARANTINE_ATTEMPT}',
    FIXTURE:            '${FIXTURE}',
    TEST:               '${TEST}',
    USERAGENT:          '${USERAGENT}',
    BROWSER:            '${BROWSER}',
    BROWSER_VERSION:    '${BROWSER_VERSION}',
    OS:                 '${OS}',
    OS_VERSION:         '${OS_VERSION}'
};

const DEFAULT_PATH_PATTERN_FOR_REPORT      = `${PLACEHOLDERS.DATE}_${PLACEHOLDERS.TIME}\\test-${PLACEHOLDERS.TEST_INDEX}`;
const DEFAULT_PATH_PATTERN                 = `${DEFAULT_PATH_PATTERN_FOR_REPORT}\\${PLACEHOLDERS.USERAGENT}\\${PLACEHOLDERS.FILE_INDEX}.${SCRENSHOT_EXTENTION}`;
const QUARANTINE_MODE_DEFAULT_PATH_PATTERN = `${DEFAULT_PATH_PATTERN_FOR_REPORT}\\run-${PLACEHOLDERS.QUARANTINE_ATTEMPT}\\${PLACEHOLDERS.USERAGENT}\\${PLACEHOLDERS.FILE_INDEX}.${SCRENSHOT_EXTENTION}`;

export default class PathPattern {
    constructor (pattern, data) {
        this.pattern              = this._ensurePattern(pattern, data.quarantineAttempt);
        this.data                 = this._addDefaultFields(data);
        this.placeholderToDataMap = this._createPlaceholderToDataMap();
    }

    _ensurePattern (pattern, quarantineAttempt) {
        if (pattern)
            return pattern;

        return quarantineAttempt ? QUARANTINE_MODE_DEFAULT_PATH_PATTERN : DEFAULT_PATH_PATTERN;
    }

    _addDefaultFields (data) {
        const defaultFields = {
            formattedDate:  data.now.format(DATE_FORMAT),
            formattedTime:  data.now.format(TIME_FORMAT),
            fileIndex:      1,
            errorFileIndex: 1
        };

        return Object.assign({}, defaultFields, data);
    }

    _createPlaceholderToDataMap () {
        return {
            [PLACEHOLDERS.DATE]:               this.data.formattedDate,
            [PLACEHOLDERS.TIME]:               this.data.formattedTime,
            [PLACEHOLDERS.TEST_INDEX]:         this.data.testIndex,
            [PLACEHOLDERS.QUARANTINE_ATTEMPT]: this.data.quarantineAttempt || 1,
            [PLACEHOLDERS.FIXTURE]:            this.data.fixture,
            [PLACEHOLDERS.TEST]:               this.data.test,
            [PLACEHOLDERS.FILE_INDEX]:         forError => forError ? this.data.errorFileIndex : this.data.fileIndex,
            [PLACEHOLDERS.USERAGENT]:          this.data.parsedUserAgent.toString(),
            [PLACEHOLDERS.BROWSER]:            this.data.parsedUserAgent.family,
            [PLACEHOLDERS.BROWSER_VERSION]:    this.data.parsedUserAgent.toVersion(),
            [PLACEHOLDERS.OS]:                 this.data.parsedUserAgent.os.family,
            [PLACEHOLDERS.OS_VERSION]:         this.data.parsedUserAgent.os.toVersion()
        };
    }

    static _buildPath (pattern, placeholderToDataMap, forError) {
        let resultFilePath = pattern;

        for (const placeholder in placeholderToDataMap) {
            const findPlaceholderRegExp = new RegExp(escapeRe(placeholder), 'g');

            resultFilePath = resultFilePath.replace(findPlaceholderRegExp, () => {
                if (placeholder === PLACEHOLDERS.FILE_INDEX) {
                    const getFileIndexFn = placeholderToDataMap[placeholder];
                    let result           = getFileIndexFn(forError);

                    if (forError)
                        result = `${ERRORS_FOLDER}\\${result}`;

                    return result;
                }

                else if (placeholder === PLACEHOLDERS.USERAGENT) {
                    const userAgent = placeholderToDataMap[placeholder];

                    return escapeUserAgent(userAgent);
                }

                return placeholderToDataMap[placeholder];
            });
        }

        return resultFilePath;
    }

    getPath (forError) {
        const path = PathPattern._buildPath(this.pattern, this.placeholderToDataMap, forError);

        return correctFilePath(path, SCRENSHOT_EXTENTION);
    }

    // For testing purposes
    static get PLACEHOLDERS () {
        return PLACEHOLDERS;
    }
}
